# -*- coding: utf-8 -*-

"""Validation tests for :mod:`biomappings`."""

import itertools as itt
import unittest
from collections import defaultdict

import bioregistry

from biomappings.resources import (
    Mappings,
    MappingTuple,
    PredictionTuple,
    load_curators,
    mapping_sort_key,
)
from biomappings.resources.semapv import get_semapv
from biomappings.utils import (
    InvalidIdentifierPattern,
    InvalidNormIdentifier,
    check_valid_prefix_id,
    get_canonical_tuple,
)

semapv = get_semapv()


def _extract_redundant(counter):
    return [(key, values) for key, values in counter.items() if len(values) > 1]


def _locations_str(locations):
    return ", ".join(f"{label}:{line}" for label, line in locations)


class IntegrityTestCase(unittest.TestCase):
    """Data integrity tests."""

    mappings: Mappings
    predictions: Mappings
    incorrect: Mappings
    unsure: Mappings

    def _iter_groups(self):
        for group, label in [
            (self.mappings, "positive"),
            (self.incorrect, "negative"),
            (self.predictions, "predictions"),
            (self.unsure, "unsure"),
        ]:
            for i, mapping in enumerate(group, start=2):
                yield label, i, mapping

    def test_prediction_types(self):
        """Test that the prediction type is pulled in properly."""
        for line, mapping in enumerate(self.mappings, start=2):
            pt = mapping.get("prediction_type", "".strip())
            if not pt:
                continue
            self.assertTrue(
                pt.startswith("semapv:"),
                msg=f"Prediction type should be annotated with semapv on line {line}",
            )
            self.assertIn(pt[len("semapv:") :], semapv)
            self.assertNotEqual(
                "semapv:ManualMappingCuration",
                pt,
                msg="Prediction can not be annotated with manual curation",
            )

        for label, line, mapping in self._iter_groups():
            tt = mapping["type"]
            self.assertTrue(
                tt.startswith("semapv:"),
                msg=f"[{label}] The 'type' column should be annotated with semapv on line {line}",
            )
            self.assertIn(tt[len("semapv:") :], semapv)

    def test_canonical_prefixes(self):
        """Test that all mappings use canonical bioregistry prefixes."""
        valid_prefixes = set(bioregistry.read_registry())
        for label, line, mapping in self._iter_groups():
            source_prefix, target_prefix = mapping["source prefix"], mapping["target prefix"]
            self.assertIn(
                source_prefix,
                valid_prefixes,
                msg=f"Invalid prefix: {source_prefix} on {label}:{line}",
            )
            self.assertIn(
                target_prefix,
                valid_prefixes,
                msg=f"Invalid prefix: {target_prefix} on {label}:{line}",
            )

    def test_normalized_identifiers(self):
        """Test that all identifiers have been normalized (based on bioregistry definition)."""
        for label, line, mapping in self._iter_groups():
            self.assert_canonical_identifier(
                mapping["source prefix"], mapping["source identifier"], label, line
            )
            self.assert_canonical_identifier(
                mapping["target prefix"], mapping["target identifier"], label, line
            )

    def assert_canonical_identifier(
        self, prefix: str, identifier: str, label: str, line: int
    ) -> None:
        """Assert a given identifier is canonical.

        :param prefix: The prefix to check
        :param identifier: The identifier in the semantic space for the prefix
        :param label: The label of the mapping file
        :param line: The line number of the mapping
        """
        try:
            check_valid_prefix_id(prefix, identifier)
        except InvalidNormIdentifier as e:
            self.fail(f"[{label}:{line}] {e}")
        except InvalidIdentifierPattern as e:
            self.fail(f"[{label}:{line}] {e}")

    def test_contributors(self):
        """Test all contributors have an entry in the curators.tsv file."""
        contributor_orcids = {row["orcid"] for row in load_curators()}
        for mapping in itt.chain(self.mappings, self.incorrect, self.unsure):
            source = mapping["source"]
            if not source.startswith("orcid:"):
                continue
            self.assertIn(source[len("orcid:") :], contributor_orcids)

    def test_cross_redundancy(self):
        """Test the redundancy of manually curated mappings and predicted mappings."""
        counter = defaultdict(lambda: defaultdict(list))
        for label, line, mapping in self._iter_groups():
            counter[get_canonical_tuple(mapping)][label].append(line)

        redundant = []
        for mapping, label_to_lines in counter.items():
            if len(label_to_lines) <= 1:
                continue
            redundant.append((mapping, sorted(label_to_lines.items())))

        if redundant:
            msg = "".join(
                f"\n  {mapping}: {_locations_str(locations)}" for mapping, locations in redundant
            )
            raise ValueError(f"{len(redundant)} are redundant: {msg}")

    def assert_no_internal_redundancies(self, m: Mappings, tuple_cls):
        """Assert that the list of mappings doesn't have any redundancies."""
        counter = defaultdict(list)
        for line, mapping in enumerate(m, start=1):
            counter[tuple_cls.from_dict(mapping)].append(line)
        redundant = _extract_redundant(counter)
        if redundant:
            msg = "".join(
                f"\n  {mapping.source_curie}/{mapping.target_curie}: {locations}"
                for mapping, locations in redundant
            )
            raise ValueError(f"{len(redundant)} are redundant: {msg}")

    def test_predictions_sorted(self):
        """Test the predictions are in a canonical order."""
        self.assertEqual(
            self.predictions,
            sorted(self.predictions, key=mapping_sort_key),
            msg="Predictions are not sorted",
        )
        self.assert_no_internal_redundancies(self.predictions, PredictionTuple)

    def test_curations_sorted(self):
        """Test the true curated mappings are in a canonical order."""
        self.assertEqual(
            self.mappings,
            sorted(self.mappings, key=mapping_sort_key),
            msg="True curations are not sorted",
        )
        self.assert_no_internal_redundancies(self.mappings, MappingTuple)

    def test_false_mappings_sorted(self):
        """Test the false curated mappings are in a canonical order."""
        self.assertEqual(
            self.incorrect,
            sorted(self.incorrect, key=mapping_sort_key),
            msg="False curations are not sorted",
        )
        self.assert_no_internal_redundancies(self.incorrect, MappingTuple)

    def test_unsure_sorted(self):
        """Test the unsure mappings are in a canonical order."""
        self.assertEqual(
            self.unsure,
            sorted(self.unsure, key=mapping_sort_key),
            msg="Unsure curations are not sorted",
        )
        self.assert_no_internal_redundancies(self.unsure, MappingTuple)



# -*- coding: utf-8 -*-

"""Validation tests for :mod:`biomappings`."""

from biomappings import (
    load_false_mappings,
    load_mappings,
    load_predictions,
    load_unsure,
    testing,
)


class TestIntegrity(testing.IntegrityTestCase):
    """Data integrity tests."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.mappings = load_mappings()
        self.predictions = load_predictions()
        self.incorrect = load_false_mappings()
        self.unsure = load_unsure()



# -*- coding: utf-8 -*-

"""Export Biomappings as SSSOM."""

import pathlib

import bioregistry
import click
import yaml
from tqdm.auto import tqdm

from biomappings import load_false_mappings, load_mappings, load_predictions
from biomappings.utils import DATA, get_curie

DIRECTORY = pathlib.Path(DATA).joinpath("sssom")
DIRECTORY.mkdir(exist_ok=True, parents=True)
TSV_PATH = DIRECTORY.joinpath("biomappings.sssom.tsv")
JSON_PATH = DIRECTORY.joinpath("biomappings.sssom.json")
OWL_PATH = DIRECTORY.joinpath("biomappings.sssom.owl")
META_PATH = DIRECTORY.joinpath("biomappings.sssom.yml")

CC0_URL = "https://creativecommons.org/publicdomain/zero/1.0/"
META = {
    "license": CC0_URL,
    "mapping_provider": "https://github.com/biopragmatics/biomappings",
    "mapping_set_group": "biomappings",
    "mapping_set_id": "biomappings",
    "mapping_set_title": "Biomappings",
}


def get_sssom_df(use_tqdm: bool = False):
    """Get an SSSOM dataframe."""
    import pandas as pd

    creators = set()

    rows = []
    prefixes = set()
    columns = [
        "subject_id",
        "subject_label",
        "predicate_id",
        "predicate_modifier",
        "object_id",
        "object_label",
        "mapping_justification",
        "author_id",
        "confidence",
        "mapping_tool",
    ]
    # see https://mapping-commons.github.io/sssom/predicate_modifier/
    # for more information on predicate modifiers
    for mappings, predicate_modifier in [
        (load_mappings(), ""),  # no predicate modifier
        (load_false_mappings(), "Not"),
    ]:
        for mapping in tqdm(mappings, unit="mapping", unit_scale=True, disable=not use_tqdm):
            prefixes.add(mapping["source prefix"])
            prefixes.add(mapping["target prefix"])
            source = mapping["source"]
            if any(source.startswith(x) for x in ["orcid:", "wikidata:"]):
                creators.add(source)
            rows.append(
                (
                    get_curie(mapping["source prefix"], mapping["source identifier"]),
                    mapping["source name"],
                    f'{mapping["relation"]}',
                    predicate_modifier,
                    get_curie(mapping["target prefix"], mapping["target identifier"]),
                    mapping["target name"],
                    mapping["type"],  # match justification
                    source,  # curator CURIE
                    mapping.get("prediction_confidence"),  # may be brought over from prediction
                    mapping.get("prediction_source"),  # may be brought over from prediction
                )
            )

    for mapping in tqdm(load_predictions(), unit="mapping", unit_scale=True, disable=not use_tqdm):
        prefixes.add(mapping["source prefix"])
        prefixes.add(mapping["target prefix"])
        rows.append(
            (
                get_curie(mapping["source prefix"], mapping["source identifier"]),
                mapping["source name"],
                f'{mapping["relation"]}',
                "",  # no predicate modifier
                get_curie(mapping["target prefix"], mapping["target identifier"]),
                mapping["target name"],
                mapping["type"],  # match justification
                None,  # no curator CURIE
                mapping["confidence"],
                mapping["source"],  # mapping tool: source script
            )
        )

    df = pd.DataFrame(rows, columns=columns)
    return prefixes, sorted(creators), df


@click.command()
def sssom():
    """Export SSSOM."""
    prefixes, creators, df = get_sssom_df()
    df.to_csv(TSV_PATH, sep="\t", index=False)

    # Get a CURIE map containing only the relevant prefixes
    prefix_map = {
        prefix: formatter
        for prefix, formatter in bioregistry.get_prefix_map().items()
        if prefix in prefixes
    }
    with open(META_PATH, "w") as file:
        yaml.safe_dump({"curie_map": prefix_map, "creator_id": creators, **META}, file)

    from sssom.parsers import from_sssom_dataframe
    from sssom.writers import write_json, write_owl

    try:
        msdf = from_sssom_dataframe(df, prefix_map=prefix_map, meta=META)
    except Exception as e:
        click.secho(f"SSSOM Export failed...\n{e}", fg="red")
        return
    click.echo("Writing JSON")
    with JSON_PATH.open("w") as file:
        write_json(msdf, file)
    click.echo("Writing OWL")
    with OWL_PATH.open("w") as file:
        write_owl(msdf, file)


if __name__ == "__main__":
    sssom()



# -*- coding: utf-8 -*-

"""Upload curation statistics to APICURON.

Run this with:

.. code-block:: sh

    $ pip install -e .[apicuron]
    $ pip install apicuron-client
    $ python -m biomappings.apicuron
"""

import datetime
from typing import Iterable

import click
from apicuron_client import Description, Report, Submission, resubmit_curations

from biomappings import load_mappings

NOW = datetime.datetime.now()
START = datetime.datetime(year=2020, month=12, day=15)

DESCRIPTION_PAYLOAD = {
    "resource_id": "Biomappings",
    "resource_name": "Biomappings",
    "resource_uri": "https://biopragmatics.github.io/biomappings/",
    "resource_url": "https://biopragmatics.github.io/biomappings/",
    "resource_long_name": "Biomappings",
    "resource_description": "Community curated and predicted equivalences and related "
    "mappings between named biological entities that are not "
    "available from primary sources.",
    "terms_def": [
        {
            "activity_term": "novel_curation",
            "activity_name": "Curated novel mapping",
            "activity_category": "generation",
            "score": 50,
            "description": "Curated a novel mapping between two entities",
        },
        {
            "activity_term": "validate_prediction",
            "activity_name": "Validate predicted mapping",
            "activity_category": "generation",
            "score": 50,
            "description": "Affirmed the correctness of a predicted mapping",
        },
        {
            "activity_term": "invalidate_prediction",
            "activity_name": "Invalidate predicted mapping",
            "activity_category": "generation",
            "score": 50,
            "description": "Affirmed the incorrectness of a predicted mapping",
        },
    ],
    "achievements_def": [
        {
            "category": "1",
            "name": "Newbie curator",
            "count_threshold": 10,
            "type": "badge",
            "list_terms": ["novel_curation", "validate_prediction", "invalidate_prediction"],
            "color_code": "#055701",
        }
    ],
}
DESCRIPTION = Description(**DESCRIPTION_PAYLOAD)


def get_curation_payload() -> Submission:
    """Get curation payload dictionary for upload to APICURON."""
    return Submission(
        resource_uri=DESCRIPTION.resource_uri,
        # time_start=START,
        # time_end=NOW,
        reports=list(iter_reports()),
    )


def iter_reports() -> Iterable[Report]:
    """Generate reports from the Biomappings for APICURON."""
    for mapping in load_mappings():
        provenance = mapping["source"]
        if not provenance.startswith("orcid:"):
            continue

        source_curie = f"{mapping['source prefix']}:{mapping['source identifier']}"
        target_curie = f"{mapping['target prefix']}:{mapping['target identifier']}"
        entity_uri = f"{source_curie}-{target_curie}"
        yield Report(
            curator_orcid=provenance[len("orcid:") :],
            activity_term="novel_curation",
            resource_uri=DESCRIPTION.resource_uri,
            # timestamp=NOW,
            entity_uri=entity_uri,
        )


@click.command()
@click.option("--token")
def main(token):
    """Submit the payload."""
    sub = get_curation_payload()
    res = resubmit_curations(sub, token=token)
    click.echo(res.text)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Load Biomappings as a graph."""

import itertools as itt
import logging
import os
from collections import Counter
from operator import itemgetter
from typing import Collection, Iterable, List, Mapping, Optional, Sequence

import click
import networkx as nx
import yaml
from bioregistry.resolve_identifier import get_bioregistry_iri
from tqdm import tqdm

from biomappings.resources import load_false_mappings, load_mappings, load_predictions
from biomappings.utils import DATA, IMG, get_curie

logger = logging.getLogger(__name__)


def get_true_graph(
    include: Optional[Sequence[str]] = None, exclude: Optional[Sequence[str]] = None
) -> nx.Graph:
    """Get a graph of the true mappings."""
    return _graph_from_mappings(load_mappings(), strata="correct", include=include, exclude=exclude)


def get_false_graph(
    include: Optional[Sequence[str]] = None, exclude: Optional[Sequence[str]] = None
) -> nx.Graph:
    """Get a graph of the false mappings."""
    return _graph_from_mappings(
        load_false_mappings(), strata="incorrect", include=include, exclude=exclude
    )


def get_predictions_graph(
    include: Optional[Collection[str]] = None, exclude: Optional[Collection[str]] = None
) -> nx.Graph:
    """Get a graph of the predicted mappings."""
    return _graph_from_mappings(
        load_predictions(), strata="predicted", include=include, exclude=exclude
    )


def _graph_from_mappings(
    mappings: Iterable[Mapping[str, str]],
    strata: str,
    include: Optional[Collection[str]] = None,
    exclude: Optional[Collection[str]] = None,
) -> nx.Graph:
    graph = nx.Graph()

    if include is not None:
        include = set(include)
        logger.info("only including %s", include)
    if exclude is not None:
        exclude = set(exclude)
        logger.info("excluding %s", exclude)

    for mapping in mappings:
        relation = mapping["relation"]
        if exclude and (relation in exclude):
            continue
        if include and (relation not in include):
            continue

        source_curie = get_curie(mapping["source prefix"], mapping["source identifier"])
        graph.add_node(
            source_curie,
            prefix=mapping["source prefix"],
            identifier=mapping["source identifier"],
            name=mapping["source name"],
        )
        target_curie = get_curie(mapping["target prefix"], mapping["target identifier"])
        graph.add_node(
            target_curie,
            prefix=mapping["target prefix"],
            identifier=mapping["target identifier"],
            name=mapping["target name"],
        )
        graph.add_edge(
            source_curie,
            target_curie,
            relation=relation,
            provenance=mapping["source"],
            type=mapping["type"],
            strata=strata,
        )
    return graph


@click.command()
def charts():
    """Make charts."""
    import matplotlib.pyplot as plt
    import seaborn as sns

    true_mappings = load_mappings()
    true_graph = _graph_from_mappings(true_mappings, include=["skos:exactMatch"], strata="correct")
    for u, v in true_graph.edges():
        true_graph.edges[u, v]["correct"] = True
    false_mappings = load_false_mappings()
    false_graph = _graph_from_mappings(
        false_mappings, include=["skos:exactMatch"], strata="incorrect"
    )
    for u, v in false_graph.edges():
        false_graph.edges[u, v]["correct"] = False

    component_node_sizes, component_edge_sizes, component_densities, component_number_prefixes = (
        [],
        [],
        [],
        [],
    )
    prefix_list = []
    components_with_duplicate_prefixes = []
    incomplete_components = []
    unstable_components = []
    n_duplicates = []
    for component in tqdm(
        nx.connected_components(true_graph), desc="Positive SCCs", unit="component", unit_scale=True
    ):
        component = true_graph.subgraph(component)
        node_size = component.number_of_nodes()
        edge_size = component.number_of_edges()

        nodes_data = {
            curie: {
                "link": get_bioregistry_iri(data["prefix"], data["identifier"]),
                **data,
            }
            for curie, data in sorted(component.nodes(data=True), key=itemgetter(0))
        }

        unstable_edges = [
            (u, v, false_graph[u, v])
            for u, v in itt.combinations(component, 2)
            if false_graph.has_edge(u, v)
        ]
        if unstable_edges:
            unstable_components.append({"nodes": nodes_data, "unstable_edges": unstable_edges})

        component_node_sizes.append(node_size)
        component_edge_sizes.append(edge_size)
        if node_size > 2:
            component_densities.append(nx.density(component))
        if node_size > 2 and edge_size < (node_size * (node_size - 1) / 2):
            incomplete_components_edges = []
            for u, v in sorted(nx.complement(component.copy()).edges()):
                if u > v:
                    u, v = v, u
                incomplete_components_edges.append(
                    {
                        "source": {"curie": u, **nodes_data[u]},
                        "target": {"curie": v, **nodes_data[v]},
                    }
                )
            incomplete_components_edges = sorted(
                incomplete_components_edges, key=lambda d: d["source"]["curie"]
            )
            incomplete_components.append(
                {
                    "nodes": nodes_data,
                    "edges": incomplete_components_edges,
                }
            )

        prefixes = [true_graph.nodes[node]["prefix"] for node in component]
        prefix_list.extend(prefixes)
        unique_prefixes = len(set(prefixes))
        component_number_prefixes.append(unique_prefixes)
        _n_duplicates = len(prefixes) - unique_prefixes
        n_duplicates.append(_n_duplicates)
        if _n_duplicates:
            components_with_duplicate_prefixes.append(nodes_data)

    with open(os.path.join(DATA, "incomplete_components.yml"), "w") as file:
        yaml.safe_dump(incomplete_components, file)
    with open(os.path.join(DATA, "components_with_duplicate_prefixes.yml"), "w") as file:
        yaml.safe_dump(components_with_duplicate_prefixes, file)
    with open(os.path.join(DATA, "unstable_components.yml"), "w") as file:
        yaml.safe_dump(unstable_components, file)

    fig, axes = plt.subplots(2, 3, figsize=(10.5, 6.5))

    _countplot_list(component_node_sizes, ax=axes[0][0])
    axes[0][0].set_yscale("log")
    axes[0][0].set_title("Size (Nodes)")

    _countplot_list(component_edge_sizes, ax=axes[0][1])
    axes[0][1].set_yscale("log")
    axes[0][1].set_title("Size (Edges)")
    axes[0][1].set_ylabel("")

    sns.kdeplot(component_densities, ax=axes[0][2])
    axes[0][2].set_xlim([0.0, 1.0])
    # axes[0][2].set_yscale('log')
    axes[0][2].set_title("Density ($|V| > 2$)")
    axes[0][2].set_ylabel("")

    _countplot_list(component_number_prefixes, ax=axes[1][0])
    axes[1][0].set_title("Number Prefixes")
    axes[1][0].set_yscale("log")
    # has duplicate prefix in component

    _countplot_list(n_duplicates, ax=axes[1][1])
    axes[1][1].set_yscale("log")
    axes[0][2].set_ylabel("")
    axes[1][1].set_title("Number Duplicate Prefixes")

    axes[1][2].axis("off")

    path = os.path.join(IMG, "components.png")
    click.echo(f"saving to {path}")
    plt.tight_layout()
    plt.savefig(path, dpi=300)
    plt.savefig(os.path.join(IMG, "components.svg"))
    plt.close(fig)

    prefix_counter = Counter(prefix_list)
    fig, axes = plt.subplots(1, 2, figsize=(8.5, 7))
    sns.countplot(y=prefix_list, ax=axes[0], order=[k for k, _ in prefix_counter.most_common()])
    axes[0].set_xscale("log")
    axes[0].set_xlabel("Count")
    axes[0].set_title(f"Prefixes ({len(prefix_counter)})")

    relations = [m["relation"] for m in true_mappings]
    relation_counter = Counter(relations)
    sns.countplot(y=relations, ax=axes[1], order=[k for k, _ in relation_counter.most_common()])
    axes[1].set_xscale("log")
    axes[1].set_xlabel("Count")
    axes[1].set_title(f"Relations ({len(relation_counter)})")

    path = os.path.join(IMG, "summary.png")
    click.echo(f"saving to {path}")
    plt.tight_layout()
    plt.savefig(path, dpi=300)
    plt.savefig(os.path.join(IMG, "summary.svg"))
    plt.close(fig)


def _countplot_list(data: List[int], ax):
    import pandas as pd
    import seaborn as sns

    counter = Counter(data)
    for size in range(min(counter), max(counter)):
        if size not in counter:
            counter[size] = 0
    df = pd.DataFrame(counter.items(), columns=["size", "count"]).sort_values("size").reset_index()
    sns.barplot(data=df, x="size", y="count", ax=ax)


if __name__ == "__main__":
    charts()



# -*- coding: utf-8 -*-

"""Biomappings resources."""

import csv
import itertools as itt
from collections import defaultdict
from pathlib import Path
from typing import (
    Any,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import bioregistry
from tqdm.auto import tqdm
from typing_extensions import Literal

from biomappings.utils import OVERRIDE_MIRIAM, RESOURCE_PATH, get_canonical_tuple

MAPPINGS_HEADER = [
    "source prefix",
    "source identifier",
    "source name",
    "relation",
    "target prefix",
    "target identifier",
    "target name",
    "type",
    "source",
    "prediction_type",
    "prediction_source",
    "prediction_confidence",
]
PREDICTIONS_HEADER = [
    "source prefix",
    "source identifier",
    "source name",
    "relation",
    "target prefix",
    "target identifier",
    "target name",
    "type",
    "confidence",
    "source",
]


class MappingTuple(NamedTuple):
    """A named tuple class for mappings."""

    source_prefix: str
    source_id: str
    source_name: str
    relation: str
    target_prefix: str
    target_identifier: str
    target_name: str
    type: str
    source: str
    prediction_type: Optional[str]
    prediction_source: Optional[str]
    prediction_confidence: Optional[float]

    def as_dict(self) -> Mapping[str, Any]:
        """Get the mapping tuple as a dictionary."""
        return dict(zip(MAPPINGS_HEADER, self))  # type:ignore

    @classmethod
    def from_dict(cls, mapping: Mapping[str, str]) -> "MappingTuple":
        """Get the mapping tuple from a dictionary."""
        values = []
        for key in MAPPINGS_HEADER:
            value = mapping.get(key) or None
            if key == "prediction_confidence" and value is not None:
                value = float(value)  # type:ignore
            values.append(value)
        return cls(*values)  # type:ignore

    @property
    def source_curie(self) -> str:
        """Concatenate the source prefix and ID to a CURIE."""
        return f"{self.source_prefix}:{self.source_id}"

    @property
    def target_curie(self) -> str:
        """Concatenate the target prefix and ID to a CURIE."""
        return f"{self.target_prefix}:{self.target_identifier}"


class PredictionTuple(NamedTuple):
    """A named tuple class for predictions."""

    source_prefix: str
    source_id: str
    source_name: str
    relation: str
    target_prefix: str
    target_identifier: str
    target_name: str
    type: str
    confidence: float
    source: str

    def as_dict(self) -> Mapping[str, Any]:
        """Get the prediction tuple as a dictionary."""
        return dict(zip(PREDICTIONS_HEADER, self))  # type:ignore

    @classmethod
    def from_dict(cls, mapping: Mapping[str, str]) -> "PredictionTuple":
        """Get the prediction tuple from a dictionary."""
        values = []
        for key in PREDICTIONS_HEADER:
            value = mapping.get(key) or None
            if key == "confidence" and value is not None:
                value = float(value)  # type:ignore
            values.append(value)
        return cls(*values)  # type:ignore

    @classmethod
    def from_semra(cls, mapping, confidence) -> "PredictionTuple":
        """Instantiate from a SeMRA mapping."""
        import pyobo
        import semra

        s_name = pyobo.get_name(*mapping.s.pair)
        if not s_name:
            raise KeyError(f"could not look up name for {mapping.s.curie}")
        o_name = pyobo.get_name(*mapping.o.pair)
        if not o_name:
            raise KeyError(f"could not look up name for {mapping.o.curie}")
        # Assume that each mapping has a single simple evidence with a mapping set annotation
        if len(mapping.evidence) != 1:
            raise ValueError
        evidence = mapping.evidence[0]
        if not isinstance(evidence, semra.SimpleEvidence):
            raise TypeError
        if evidence.mapping_set is None:
            raise ValueError
        return cls(  # type:ignore
            *mapping.s.pair,
            s_name,
            mapping.p.curie,
            *mapping.o.pair,
            o_name,
            evidence.justification.curie,
            confidence,
            evidence.mapping_set.name,
        )

    @property
    def source_curie(self) -> str:
        """Concatenate the source prefix and ID to a CURIE."""
        return f"{self.source_prefix}:{self.source_id}"

    @property
    def target_curie(self) -> str:
        """Concatenate the target prefix and ID to a CURIE."""
        return f"{self.target_prefix}:{self.target_identifier}"


Mappings = Iterable[Mapping[str, str]]


def get_resource_file_path(fname) -> Path:
    """Get a resource by its file name."""
    return RESOURCE_PATH.joinpath(fname)


def _load_table(fname) -> List[Dict[str, str]]:
    with open(fname, "r") as fh:
        reader = csv.reader(fh, delimiter="\t")
        header = next(reader)
        return [_clean(header, row) for row in reader]


def _clean(header, row):
    d = dict(zip(header, row))
    return {k: v if v else None for k, v in d.items()}


def _write_helper(
    header: Sequence[str], mappings: Mappings, path: Union[str, Path], mode: Literal["w", "a"]
) -> None:
    mappings = sorted(mappings, key=mapping_sort_key)
    with open(path, mode) as file:
        if mode == "w":
            print(*header, sep="\t", file=file)  # noqa:T201
        for line in mappings:
            print(*[line[k] or "" for k in header], sep="\t", file=file)  # noqa:T201


def mapping_sort_key(prediction: Mapping[str, str]) -> Tuple[str, ...]:
    """Return a tuple for sorting mapping dictionaries."""
    return (
        prediction["source prefix"],
        prediction["source identifier"],
        prediction["relation"],
        prediction["target prefix"],
        prediction["target identifier"],
        prediction["type"],
        prediction["source"] or "",
    )


TRUE_MAPPINGS_PATH = get_resource_file_path("mappings.tsv")


def load_mappings(*, path: Optional[Path] = None) -> List[Dict[str, str]]:
    """Load the mappings table."""
    return _load_table(path or TRUE_MAPPINGS_PATH)


def append_true_mappings(
    mappings: Mappings,
    *,
    sort: bool = True,
    path: Optional[Path] = None,
) -> None:
    """Append new lines to the mappings table."""
    if path is None:
        path = TRUE_MAPPINGS_PATH
    _write_curated(mappings, path=path, mode="a")
    if sort:
        lint_true_mappings(path=path)


def append_true_mapping_tuples(mappings: Iterable[MappingTuple]) -> None:
    """Append new lines to the mappings table."""
    append_true_mappings(mapping.as_dict() for mapping in set(mappings))


def write_true_mappings(mappings: Mappings, *, path: Optional[Path] = None) -> None:
    """Write mappigns to the true mappings file."""
    _write_curated(mappings=mappings, path=path or TRUE_MAPPINGS_PATH, mode="w")


def _write_curated(mappings: Mappings, *, path: Path, mode: Literal["w", "a"]):
    _write_helper(MAPPINGS_HEADER, mappings, path, mode=mode)


def lint_true_mappings(*, standardize: bool = False, path: Optional[Path] = None) -> None:
    """Lint the true mappings file."""
    _lint_curated_mappings(standardize=standardize, path=path or TRUE_MAPPINGS_PATH)


def _lint_curated_mappings(path: Path, *, standardize: bool = False) -> None:
    """Lint the true mappings file."""
    mappings = _load_table(path)
    mappings = _remove_redundant(mappings, standardize=standardize)
    _write_helper(MAPPINGS_HEADER, mappings, path, mode="w")


FALSE_MAPPINGS_PATH = get_resource_file_path("incorrect.tsv")


def load_false_mappings(*, path: Optional[Path] = None) -> List[Dict[str, str]]:
    """Load the false mappings table."""
    return _load_table(path or FALSE_MAPPINGS_PATH)


def append_false_mappings(
    mappings: Mappings,
    *,
    sort: bool = True,
    path: Optional[Path] = None,
) -> None:
    """Append new lines to the false mappings table."""
    if path is None:
        path = FALSE_MAPPINGS_PATH
    _write_curated(mappings=mappings, path=path, mode="a")
    if sort:
        lint_false_mappings(path=path)


def write_false_mappings(mappings: Mappings, *, path: Optional[Path] = None) -> None:
    """Write mappings to the false mappings file."""
    _write_helper(MAPPINGS_HEADER, mappings, path or FALSE_MAPPINGS_PATH, mode="w")


def lint_false_mappings(*, standardize: bool = False, path: Optional[Path] = None) -> None:
    """Lint the false mappings file."""
    _lint_curated_mappings(standardize=standardize, path=path or FALSE_MAPPINGS_PATH)


UNSURE_PATH = get_resource_file_path("unsure.tsv")


def load_unsure(*, path: Optional[Path] = None) -> List[Dict[str, str]]:
    """Load the unsure table."""
    return _load_table(path or UNSURE_PATH)


def append_unsure_mappings(
    mappings: Mappings,
    *,
    sort: bool = True,
    path: Optional[Path] = None,
) -> None:
    """Append new lines to the "unsure" mappings table."""
    if path is None:
        path = UNSURE_PATH
    _write_curated(mappings, path=path, mode="a")
    if sort:
        lint_unsure_mappings(path=path)


def write_unsure_mappings(mappings: Mappings, *, path: Optional[Path] = None) -> None:
    """Write mappings to the unsure mappings file."""
    _write_helper(MAPPINGS_HEADER, mappings, path or UNSURE_PATH, mode="w")


def lint_unsure_mappings(*, standardize: bool = False, path: Optional[Path] = None) -> None:
    """Lint the unsure mappings file."""
    _lint_curated_mappings(standardize=standardize, path=path or UNSURE_PATH)


PREDICTIONS_PATH = get_resource_file_path("predictions.tsv")


def load_predictions(*, path: Optional[Path] = None) -> List[Dict[str, str]]:
    """Load the predictions table."""
    return _load_table(path or PREDICTIONS_PATH)


def write_predictions(mappings: Mappings, *, path: Optional[Path] = None) -> None:
    """Write new content to the predictions table."""
    _write_helper(PREDICTIONS_HEADER, mappings, path or PREDICTIONS_PATH, mode="w")


def append_prediction_tuples(
    prediction_tuples: Iterable[PredictionTuple],
    *,
    deduplicate: bool = True,
    sort: bool = True,
    standardize: bool = True,
) -> None:
    """Append new lines to the predictions table that come as canonical tuples."""
    append_predictions(
        (prediction_tuple.as_dict() for prediction_tuple in set(prediction_tuples)),
        deduplicate=deduplicate,
        sort=sort,
        standardize=standardize,
    )


def append_predictions(
    mappings: Mappings,
    *,
    deduplicate: bool = True,
    sort: bool = True,
    standardize: bool = True,
) -> None:
    """Append new lines to the predictions table."""
    if standardize:
        mappings = _standardize_mappings(mappings)
    if deduplicate:
        existing_mappings = {
            get_canonical_tuple(existing_mapping)
            for existing_mapping in itt.chain(
                load_mappings(),
                load_false_mappings(),
                load_unsure(),
                load_predictions(),
            )
        }
        mappings = (
            mapping for mapping in mappings if get_canonical_tuple(mapping) not in existing_mappings
        )

    _write_helper(PREDICTIONS_HEADER, mappings, PREDICTIONS_PATH, mode="a")
    if sort:
        lint_predictions()


def lint_predictions(standardize: bool = False) -> None:
    """Lint the predictions file.

    1. Make sure there are no redundant rows
    2. Make sure no rows in predictions match a row in the curated files
    3. Make sure it's sorted

    :param standardize: Should identifiers be standardized (against the
             combination of Identifiers.org and Bioregistry)?
    """
    curated_mappings = {
        get_canonical_tuple(mapping)
        for mapping in itt.chain(
            load_mappings(),
            load_false_mappings(),
            load_unsure(),
        )
    }
    mappings = [
        mapping
        for mapping in tqdm(
            load_predictions(), desc="Removing curated from predicted", unit_scale=True
        )
        if get_canonical_tuple(mapping) not in curated_mappings
    ]
    mappings = _remove_redundant(mappings, standardize=standardize)
    mappings = sorted(mappings, key=mapping_sort_key)
    write_predictions(mappings)


def _remove_redundant(mappings, *, standardize: bool = False):
    if standardize:
        mappings = _standardize_mappings(mappings)
    dd = defaultdict(list)
    for mapping in mappings:
        dd[get_canonical_tuple(mapping)].append(mapping)
    return [max(mappings, key=_pick_best) for mappings in dd.values()]


def _pick_best(mapping: Dict[str, str]) -> int:
    """Assign a value for this mapping.

    :param mapping: A mapping dictionary
    :returns: An integer, where higher means a better choice.

    This function is currently simple, but can later be extended to
    account for several other things including:

    - confidence in the curator
    - prediction methodology
    - date of prediction/curation (to keep the earliest)
    """
    if mapping["source"].startswith("orcid"):
        return 1
    return 0


def _standardize_mappings(mappings, *, progress: bool = True):
    for mapping in tqdm(
        mappings,
        desc="Standardizing mappings",
        unit_scale=True,
        unit="mapping",
        disable=not progress,
    ):
        yield _standardize_mapping(mapping)


def _standardize_mapping(mapping):
    """Standardize a mapping."""
    for prefix_key, identifier_key in [
        ("source prefix", "source identifier"),
        ("target prefix", "target identifier"),
    ]:
        prefix, identifier = mapping[prefix_key], mapping[identifier_key]
        resource = bioregistry.get_resource(prefix)
        if resource is None:
            raise ValueError
        miriam_prefix = resource.get_miriam_prefix()
        if miriam_prefix is None or miriam_prefix in OVERRIDE_MIRIAM:
            mapping[identifier_key] = resource.standardize_identifier(identifier)
        else:
            mapping[identifier_key] = (
                resource.miriam_standardize_identifier(identifier) or identifier
            )

    return mapping


def load_curators():
    """Load the curators table."""
    return _load_table(get_resource_file_path("curators.tsv"))


def filter_predictions(custom_filter: Mapping[str, Mapping[str, Mapping[str, str]]]) -> None:
    """Filter all the predictions by removing what's in the custom filter then re-write.

    :param custom_filter: A filter 3-dictionary of source prefix to target prefix
        to source identifier to target identifier
    """
    predictions = load_predictions()
    predictions = [
        prediction for prediction in predictions if _check_filter(prediction, custom_filter)
    ]
    write_predictions(predictions)


def _check_filter(
    prediction: Mapping[str, str],
    custom_filter: Mapping[str, Mapping[str, Mapping[str, str]]],
) -> bool:
    source_prefix, target_prefix = prediction["source prefix"], prediction["target prefix"]
    source_id, target_id = prediction["source identifier"], prediction["target identifier"]
    return target_id != custom_filter.get(source_prefix, {}).get(target_prefix, {}).get(source_id)


def get_curated_filter() -> Mapping[str, Mapping[str, Mapping[str, str]]]:
    """Get a filter over all curated mappings."""
    d: DefaultDict[str, DefaultDict[str, Dict[str, str]]] = defaultdict(lambda: defaultdict(dict))
    for m in itt.chain(load_mappings(), load_false_mappings(), load_unsure()):
        d[m["source prefix"]][m["target prefix"]][m["source identifier"]] = m["target identifier"]
    return {k: dict(v) for k, v in d.items()}


def prediction_tuples_from_semra(
    mappings,
    *,
    confidence: float,
) -> List[PredictionTuple]:
    """Get prediction tuples from SeMRA mappings."""
    rows = []
    for mapping in mappings:
        try:
            row = PredictionTuple.from_semra(mapping, confidence)
        except KeyError as e:
            tqdm.write(str(e))
            continue
        rows.append(row)
    return rows



"""Tools for semapv."""

import json
from functools import lru_cache
from pathlib import Path

__all__ = [
    "get_semapv",
]

url = "https://raw.githubusercontent.com/mapping-commons/semantic-mapping-vocabulary/main/semapv-terms.tsv"
HERE = Path(__file__).parent.resolve()
PATH = HERE.joinpath("semapv.json")


@lru_cache(1)
def get_semapv():
    """Get a dictionary of semapv local unique identifier to labels."""
    if PATH.is_file():
        return json.loads(PATH.read_text())

    import pystow

    df = pystow.ensure_csv(
        "bio",
        "semapv",
        url=url,
        read_csv_kwargs=dict(
            usecols=[0, 1],
            skiprows=1,
        ),
    )
    df["ID"] = df["ID"].map(lambda s: s.removeprefix("semapv:"))
    rv = dict(df.values)
    PATH.write_text(json.dumps(rv, indent=True, sort_keys=True))
    return rv


if __name__ == "__main__":
    get_semapv()



"""Code for the paper analysis."""

import json
import pickle
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import DefaultDict, Dict, Iterable, Mapping, Tuple

import bioontologies
import bioregistry
import pyobo
import pystow
from bioontologies.obograph import _parse_uri_or_curie_or_str
from bioregistry import manager
from tabulate import tabulate
from tqdm.auto import tqdm

__all__ = [
    "Result",
    "get_primary_mappings",
    "get_obo_mappings",
    "get_non_obo_mappings",
    "index_mappings",
]

EVALUATION = pystow.module("biomappings", "evaluation")


@dataclass
class Result:
    """Added value result set."""

    dataset: str
    source: str
    target: str
    total: int
    missing: int
    missing_biomappings: int
    missing_predictions: int

    @classmethod
    def make(
        cls,
        dataset,
        source,
        target,
        datasource_identifiers,
        primary,
        secondary,
        tertiary,
    ):
        """Create a value added summary object with analysis over several dictionaries."""
        return cls._from_dicts(
            dataset=dataset,
            source=source,
            target=target,
            datasource_identifiers=datasource_identifiers,
            ontology_external_identifiers=set(primary.get(source, {}).get(target, {})),
            biomappings_external_identifiers=set(secondary.get(source, {}).get(target, {})),
            biomappings_prediction_identifiers=set(tertiary.get(source, {}).get(target, {})),
        )

    @classmethod
    def _from_dicts(
        cls,
        dataset,
        source,
        target,
        datasource_identifiers,
        ontology_external_identifiers,
        biomappings_external_identifiers,
        biomappings_prediction_identifiers,
    ):
        """Create a value added summary object with analysis over several dictionaries."""
        return Result(
            dataset=dataset,
            source=source,
            target=target,
            total=len(datasource_identifiers),
            missing=len(datasource_identifiers - ontology_external_identifiers),
            missing_biomappings=len(
                datasource_identifiers
                - ontology_external_identifiers
                - biomappings_external_identifiers
            ),
            missing_predictions=len(
                datasource_identifiers
                - ontology_external_identifiers
                - biomappings_external_identifiers
                - biomappings_prediction_identifiers
            ),
        )

    def print(self):  # noqa:T202
        """Print a summary of value added statistics."""
        print(  # noqa:T201
            tabulate(
                [
                    (f"Total in {self.dataset}", f"{self.total:,}", ""),
                    (
                        f"Missing w/ {self.source}",
                        f"{self.missing:,}",
                        f"{self.missing / self.total:.2%}",
                    ),
                    (
                        f"Missing w/ {self.source} + BM.",
                        f"{self.missing_biomappings:,}",
                        f"{self.missing_biomappings / self.total:.1%}",
                    ),
                    (
                        f"Missing w/ {self.source} + BM. + Pred.",
                        f"{self.missing_predictions:,}",
                        f"{self.missing_predictions / self.total:.1%}",
                    ),
                ],
                headers=["Missing", f"Unmappable to {self.target}", "% Unmappable"],
            )
        )


def get_primary_mappings(
    prefix: str,
    external_prefix: str,
    cache_path: Path,
) -> Tuple[str, Mapping[str, str]]:
    """Get mappings from a given ontology (prefix) to another resource (external prefix)."""
    if cache_path.is_file():
        d = json.loads(cache_path.read_text())
        return d["version"], d["mappings"]

    parse_results = bioontologies.get_obograph_by_prefix(prefix)
    version = parse_results.guess_version(prefix)
    graphs = parse_results.graph_document.graphs if parse_results.graph_document else []
    rv: Dict[str, str] = {}
    for graph in graphs:
        for node in tqdm(
            graph.nodes,
            unit="node",
            unit_scale=True,
            leave=False,
            desc=f"Extracting {external_prefix} from {prefix}",
        ):
            try:
                node_prefix, node_luid = _parse_uri_or_curie_or_str(node.id)
            except ValueError:
                continue

            if node_prefix is None or bioregistry.normalize_prefix(node_prefix) != prefix:
                continue

            for xref in node.xrefs:
                xref_prefix, xref_luid = bioregistry.parse_curie(xref.val)
                if xref_prefix != external_prefix:
                    continue
                rv[xref_luid] = node_luid

    d = {"mappings": rv, "version": version}
    cache_path.write_text(json.dumps(d, indent=2, sort_keys=True))
    return version, rv


def index_mappings(mappings: Iterable[Mapping[str, str]], path=None, force: bool = False):
    """Create an index of mappings."""
    if path and path.is_file() and not force:
        with open(path, "rb") as file:
            return pickle.load(file)

    rv: DefaultDict[str, DefaultDict[str, Dict[str, str]]] = defaultdict(lambda: defaultdict(dict))

    for mapping in tqdm(mappings, unit_scale=True, unit="mapping"):
        source_prefix = mapping["source prefix"]
        source_resource = manager.registry[source_prefix]
        source_id = source_resource.standardize_identifier(mapping["source identifier"])
        target_prefix = mapping["target prefix"]
        target_resource = manager.registry[target_prefix]
        target_id = target_resource.standardize_identifier(mapping["target identifier"])
        rv[source_prefix][target_prefix][source_id] = target_id
        rv[target_prefix][source_prefix][target_id] = source_id

    rvp = {k: dict(v) for k, v in rv.items()}
    if path:
        with open(path, "wb") as file:
            pickle.dump(rvp, file)
    return rvp


PRIMARY_MAPPING_CONFIG = [
    ("doid", "umls", "http://purl.obolibrary.org/obo/doid.owl"),
    ("doid", "mesh", "http://purl.obolibrary.org/obo/doid.owl"),
    ("doid", "mondo", "http://purl.obolibrary.org/obo/doid.owl"),
    ("doid", "efo", "http://purl.obolibrary.org/obo/doid.owl"),
    ("mondo", "umls", "http://purl.obolibrary.org/obo/mondo.owl"),
    ("mondo", "mesh", "http://purl.obolibrary.org/obo/mondo.owl"),
    ("mondo", "doid", "http://purl.obolibrary.org/obo/mondo.owl"),
    ("mondo", "efo", "http://purl.obolibrary.org/obo/mondo.owl"),
    ("efo", "mesh", "http://www.ebi.ac.uk/efo/efo.obo"),
    ("efo", "doid", "http://www.ebi.ac.uk/efo/efo.obo"),
    ("efo", "cl", "http://www.ebi.ac.uk/efo/efo.obo"),
    ("efo", "ccle", "http://www.ebi.ac.uk/efo/efo.obo"),
    ("hp", "mesh", "http://purl.obolibrary.org/obo/hp.owl"),
    ("go", "mesh", "http://purl.obolibrary.org/obo/go.owl"),
    ("go", "reactome", "http://purl.obolibrary.org/obo/go.owl"),
    ("go", "wikipathways", "http://purl.obolibrary.org/obo/go.owl"),
    # ("clo", "efo", "http://purl.obolibrary.org/obo/clo.owl"),
    # ("clo", "ccle", "http://purl.obolibrary.org/obo/clo.owl"),
    # ("clo", "depmap", "http://purl.obolibrary.org/obo/clo.owl"),
    # ("clo", "cellosaurus", "http://purl.obolibrary.org/obo/clo.owl"),
    ("uberon", "mesh", "http://purl.obolibrary.org/obo/uberon.owl"),
    ("cl", "efo", "http://purl.obolibrary.org/obo/cl.owl"),
    ("cl", "mesh", "http://purl.obolibrary.org/obo/cl.owl"),
    ("chebi", "mesh", "http://purl.obolibrary.org/obo/chebi.owl"),
    ("chebi", "ncit", "http://purl.obolibrary.org/obo/chebi.owl"),
]


def _clean_version(version: str, prefix) -> str:
    return (
        version.removeprefix(f"http://purl.obolibrary.org/obo/{prefix}/")
        .removeprefix("releases/")
        .removesuffix(f"/{prefix}.owl")
    )


def get_obo_mappings(primary_dd, biomappings_dd):
    """Fill the primary mappings for ontologies and calculate a value added summary."""
    summary_rows = []
    for prefix, external, _uri in PRIMARY_MAPPING_CONFIG:
        cache_path = EVALUATION.join("mappings", name=f"{prefix}_{external}.json")
        version, primary = get_primary_mappings(prefix, external, cache_path=cache_path)
        primary_dd[external][prefix] = primary

        primary_set = set(primary)
        bm = set(biomappings_dd.get(external, {}).get(prefix, {})).union(
            biomappings_dd.get(prefix, {}).get(external, {})
        )
        n_primary = len(primary_set.difference(bm))
        n_biomappings = len(bm)
        n_total = len(primary_set.union(bm))

        if not n_primary and n_biomappings:
            gain = float("inf")
        elif not n_primary and not n_biomappings:
            continue  # skip this
            # gain = "-"
        else:
            gain = round(100 * n_biomappings / n_primary, 1) if n_primary else None

        summary_rows.append(
            (
                prefix,
                _clean_version(version, prefix),
                external,
                n_primary,
                n_biomappings,
                n_total,
                gain,
            )
        )
    return summary_rows


PYOBO_CONFIGS = [
    ("cellosaurus", "efo", "CVCL_", "EFO_"),
    ("cellosaurus", "ccle", "CVCL_", ""),
    # ("cellosaurus", "cl", "CVCL_", ""),
]


def get_non_obo_mappings(primary_dd, biomappings_dd):
    """Fill the primary mappings for non-obo sources calculate a value added summary."""
    summary_rows = []
    for prefix, external, source_banana, target_banana in PYOBO_CONFIGS:
        xrefs_df = pyobo.get_xrefs_df(prefix)
        xrefs_slim_df = xrefs_df[xrefs_df["target_ns"] == external]

        version = "unknown"  # FIXME, e.g., with bioversions.get_version(prefix)
        primary = primary_dd[external][prefix] = {
            target_id.removeprefix(target_banana): source_id.removeprefix(source_banana)
            for source_id, target_id in xrefs_slim_df[[f"{prefix}_id", "target_id"]].values
        }
        n_primary = len(primary)

        bm = set(biomappings_dd.get(external, {}).get(prefix, {})).union(
            biomappings_dd.get(prefix, {}).get(external, {})
        )
        n_biomappings = len(bm)
        n_total = len(set(primary).union(bm))

        if not n_primary and n_biomappings:
            gain = float("inf")
        elif not n_primary and not n_biomappings:
            gain = "-"
        else:
            gain = round(100 * n_biomappings / n_primary, 1) if n_primary else None

        summary_rows.append(
            (
                prefix,
                _clean_version(version, prefix),
                external,
                n_primary,
                n_biomappings,
                n_total,
                gain,
            )
        )
    return summary_rows



# -*- coding: utf-8 -*-

"""Functions for working with the mapping graph."""

import itertools as itt
from collections import defaultdict
from typing import TYPE_CHECKING, DefaultDict, Dict, Iterable, List, Optional

import networkx as nx
import pyobo

from biomappings.utils import CMapping

if TYPE_CHECKING:
    import semra

__all__ = [
    "get_custom_filter",
    "get_filter_from_semra",
    "mutual_mapping_graph",
]


def get_filter_from_semra(mappings: List["semra.Mapping"]) -> CMapping:
    """Get a custom filter dictionary from a set of SeMRA mappings."""
    rv: DefaultDict[str, DefaultDict[str, Dict[str, str]]] = defaultdict(lambda: defaultdict(dict))
    for mapping in mappings:
        rv[mapping.s.prefix][mapping.o.prefix][mapping.s.identifier] = mapping.o.identifier
    return rv


def get_custom_filter(prefix: str, targets: Iterable[str]) -> CMapping:
    """Get a custom filter dictionary induced over the mutual mapping graph with all target prefixes.

    :param prefix: The source prefix
    :param targets: All potential target prefixes
    :returns: A filter 3-dictionary of source prefix to target prefix to source identifier to target identifier
    """
    graph = mutual_mapping_graph([prefix, *targets])
    rv: DefaultDict[str, Dict[str, str]] = defaultdict(dict)
    for p, identifier in graph:
        if p != prefix:
            continue
        for xref_prefix, xref_identifier in nx.single_source_shortest_path(graph, (p, identifier)):
            rv[xref_prefix][identifier] = xref_identifier
    return {prefix: dict(rv)}


def mutual_mapping_graph(
    prefixes: Iterable[str],
    skip_sources: Optional[Iterable[str]] = None,
    skip_targets: Optional[Iterable[str]] = None,
) -> nx.Graph:
    """Get the undirected mapping graph between the given prefixes.

    :param prefixes: A list of prefixes to use with :func:`pyobo.get_filtered_xrefs` to get xrefs.
    :param skip_sources: An optional list of prefixes to skip as the source for xrefs
    :param skip_targets: An optional list of prefixes to skip as the target for xrefs
    :return: The undirected mapping graph containing mappings between entries in the given namespaces.
    """
    prefixes = sorted(prefixes)
    skip_sources = set() if skip_sources is None else set(skip_sources)
    skip_targets = set() if skip_targets is None else set(skip_targets)
    graph = nx.Graph()
    for source, target in itt.product(prefixes, repeat=2):
        if source == target or source in skip_sources or target in skip_targets:
            continue
        for source_id, target_id in pyobo.get_filtered_xrefs(source, target).items():
            graph.add_edge((source, source_id), (target, target_id))
    return graph



# -*- coding: utf-8 -*-

"""Utilities for generating predictions with pyobo/gilda."""

import logging
from collections import defaultdict
from typing import Iterable, Optional, Tuple, Union

import bioregistry
import pyobo
from gilda.grounder import Grounder
from pyobo import get_xref
from pyobo.gilda_utils import get_grounder, iter_gilda_prediction_tuples

from biomappings.resources import PredictionTuple, append_prediction_tuples
from biomappings.utils import CMapping

logger = logging.getLogger(__name__)


def append_gilda_predictions(
    prefix: str,
    target_prefixes: Union[str, Iterable[str]],
    provenance: str,
    *,
    relation: str = "skos:exactMatch",
    custom_filter: Optional[CMapping] = None,
    unnamed: Optional[Iterable[str]] = None,
    identifiers_are_names: bool = False,
) -> None:
    """Add gilda predictions to the Biomappings predictions.tsv file.

    :param prefix: The source prefix
    :param target_prefixes: The target prefix or prefixes
    :param provenance: The provenance text. Typically generated with ``biomappings.utils.get_script_url(__file__)``.
    :param relation: The relationship. Defaults to ``skos:exactMatch``.
    :param custom_filter: A triple nested dictionary from source prefix to target prefix to source id to target id.
        Any source prefix, target prefix, source id combinations in this dictionary will be filtered.
    :param unnamed: An optional list of prefixes whose identifiers should be considered as names (e.g., CCLE, FPLX)
    :param identifiers_are_names: The source prefix's identifiers should be considered as names
    """
    if isinstance(target_prefixes, str):
        target_prefixes = [target_prefixes]
    grounder = get_grounder(target_prefixes, unnamed=unnamed)
    predictions = iter_prediction_tuples(
        prefix,
        relation=relation,
        grounder=grounder,
        provenance=provenance,
        identifiers_are_names=identifiers_are_names,
    )
    if custom_filter is not None:
        predictions = filter_custom(predictions, custom_filter)
    predictions = filter_existing_xrefs(predictions, [prefix, *target_prefixes])
    predictions = sorted(predictions, key=_key)
    append_prediction_tuples(predictions)


def iter_prediction_tuples(
    prefix: str,
    provenance: str,
    *,
    relation: str = "skos:exactMatch",
    grounder: Optional[Grounder] = None,
    identifiers_are_names: bool = False,
) -> Iterable[PredictionTuple]:
    """Iterate over prediction tuples for a given prefix."""
    for t in iter_gilda_prediction_tuples(
        prefix=prefix,
        relation=relation,
        grounder=grounder,
        identifiers_are_names=identifiers_are_names,
    ):
        yield PredictionTuple(*t, provenance)  # type: ignore


def filter_custom(
    predictions: Iterable[PredictionTuple],
    custom_filter: CMapping,
) -> Iterable[PredictionTuple]:
    """Filter out custom mappings."""
    counter = 0
    for p in predictions:
        if custom_filter.get(p.source_prefix, {}).get(p.target_prefix, {}).get(p.source_id):
            counter += 1
            continue
        yield p
    logger.info("filtered out %d custom mapped matches", counter)


def filter_existing_xrefs(
    predictions: Iterable[PredictionTuple], prefixes: Iterable[str]
) -> Iterable[PredictionTuple]:
    """Filter predictions that match xrefs already loaded through PyOBO."""
    prefixes = set(prefixes)

    entity_to_mapped_prefixes = defaultdict(set)
    for prefix in prefixes:
        for source_id, target_prefix, target_id in pyobo.get_xrefs_df(prefix).values:
            entity_to_mapped_prefixes[prefix, source_id].add(target_prefix)
            entity_to_mapped_prefixes[target_prefix, target_id].add(prefix)

    counter = 0
    for prediction in predictions:
        source_id = bioregistry.standardize_identifier(
            prediction.source_prefix, prediction.source_id
        )
        target_id = bioregistry.standardize_identifier(
            prediction.target_prefix, prediction.target_identifier
        )
        if (
            prediction.target_prefix
            in entity_to_mapped_prefixes[prediction.source_prefix, source_id]
            or prediction.source_prefix
            in entity_to_mapped_prefixes[prediction.target_prefix, target_id]
        ):
            counter += 1
            continue
        yield prediction
    logger.info("filtered out %d pre-mapped matches", counter)


def has_mapping(prefix: str, identifier: str, target_prefix: str) -> bool:
    """Check if there's already a mapping available for this entity in a target namespace."""
    return get_xref(prefix, identifier, target_prefix) is not None


def _key(t: PredictionTuple) -> Tuple[str, str]:
    return t.source_prefix, t.source_name
